@@ -41,7 +41,7 @@
 #include <mntent.h>
 #include <fcntl.h>
 #include <limits.h>
-#include <fstab.h>
+#include <paths.h>
 #include <sys/mman.h>
 #include <sys/wait.h>
 #ifdef HAVE_LIBCAP_NG
@@ -56,6 +56,7 @@
 #endif /* HAVE_LIBCAP_NG */
 #include "mount.h"
 #include "util.h"
+#include "resolve_host.h"
 
 #ifndef MS_MOVE 
 #define MS_MOVE 8192 
@@ -87,12 +88,6 @@
 /* max length of username (somewhat made up here) */
 #define MAX_USERNAME_SIZE 32
 
-/* currently maximum length of IPv6 address string */
-#define MAX_ADDRESS_LEN INET6_ADDRSTRLEN
-
-/* limit list of addresses to 16 max-size addrs */
-#define MAX_ADDR_LIST_LEN ((MAX_ADDRESS_LEN + 1) * 16)
-
 #ifndef SAFE_FREE
 #define SAFE_FREE(x) do { if ((x) != NULL) {free(x); x = NULL; } } while (0)
 #endif
@@ -133,6 +128,7 @@
  * Values for parsing command line options.
  */
 #define OPT_ERROR       -1
+#define OPT_IGNORE      0
 #define OPT_USERS       1
 #define OPT_USER        2
 #define OPT_USER_XATTR  3
@@ -159,6 +155,9 @@
 #define OPT_RO         24
 #define OPT_RW         25
 #define OPT_REMOUNT    26
+#define OPT_MAND       27
+#define OPT_NOMAND     28
+#define OPT_CRUID      29
 
 
 /* struct for holding parsed mount info for use by privleged process */
@@ -212,9 +211,9 @@ check_fstab(const char *progname, const char *mountpoint, const char *devname,
 	struct mntent *mnt;
 
 	/* make sure this mount is listed in /etc/fstab */
-	fstab = setmntent(_PATH_FSTAB, "r");
+	fstab = setmntent(_PATH_MNTTAB, "r");
 	if (!fstab) {
-		fprintf(stderr, "Couldn't open %s for reading!\n", _PATH_FSTAB);
+		fprintf(stderr, "Couldn't open %s for reading!\n", _PATH_MNTTAB);
 		return EX_FILEIO;
 	}
 
@@ -226,7 +225,7 @@ check_fstab(const char *progname, const char *mountpoint, const char *devname,
 
 	if (mnt == NULL || strcmp(mnt->mnt_fsname, devname)) {
 		fprintf(stderr, "%s: permission denied: no match for "
-			"%s found in %s\n", progname, mountpoint, _PATH_FSTAB);
+			"%s found in %s\n", progname, mountpoint, _PATH_MNTTAB);
 		return EX_USAGE;
 	}
 
@@ -265,7 +264,7 @@ static int mount_cifs_usage(FILE * stream)
 	fprintf(stream,
 		"\n\tmapchars,nomapchars,nolock,servernetbiosname=<SRV_RFC1001NAME>");
 	fprintf(stream,
-		"\n\tdirectio,nounix,cifsacl,sec=<authentication mechanism>,sign");
+		"\n\tdirectio,nounix,cifsacl,sec=<authentication mechanism>,sign,fsc");
 	fprintf(stream,
 		"\n\nOptions not needed for servers supporting CIFS Unix extensions");
 	fprintf(stream,
@@ -637,7 +636,7 @@ parsing_err:
 static int open_cred_file(char *file_name,
 			struct parsed_mount_info *parsed_info)
 {
-	char *line_buf;
+	char *line_buf = NULL;
 	char *temp_val = NULL;
 	FILE *fs = NULL;
 	int i;
@@ -800,8 +799,12 @@ static int parse_opt_token(const char *token)
 		return OPT_UNC;
 	if (strncmp(token, "dom", 3) == 0 || strncmp(token, "workg", 5) == 0)
 		return OPT_DOM;
+	if (strncmp(token, "cred", 4) == 0)
+		return OPT_CRED;
 	if (strncmp(token, "uid", 3) == 0)
 		return OPT_UID;
+	if (strncmp(token, "cruid", 5) == 0)
+		return OPT_CRUID;
 	if (strncmp(token, "gid", 3) == 0)
 		return OPT_GID;
 	if (strncmp(token, "fmask", 5) == 0)
@@ -810,7 +813,7 @@ static int parse_opt_token(const char *token)
 		return OPT_FILE_MODE;
 	if (strncmp(token, "dmask", 5) == 0)
 		return OPT_DMASK;
-	if (strncmp(token, "dir_mode", 8) == 0)
+	if (strncmp(token, "dir_mode", 4) == 0 || strncmp(token, "dirm", 4) == 0)
 		return OPT_DIR_MODE;
 	if (strncmp(token, "nosuid", 6) == 0)
 		return OPT_NO_SUID;
@@ -820,6 +823,10 @@ static int parse_opt_token(const char *token)
 		return OPT_NO_DEV;
 	if (strncmp(token, "nobrl", 5) == 0 || strncmp(token, "nolock", 6) == 0)
 		return OPT_NO_LOCK;
+	if (strncmp(token, "mand", 4) == 0)
+		return OPT_MAND;
+	if (strncmp(token, "nomand", 6) == 0)
+		return OPT_NOMAND;
 	if (strncmp(token, "dev", 3) == 0)
 		return OPT_DEV;
 	if (strncmp(token, "noexec", 6) == 0)
@@ -834,6 +841,8 @@ static int parse_opt_token(const char *token)
 		return OPT_RW;
 	if (strncmp(token, "remount", 7) == 0)
 		return OPT_REMOUNT;
+	if (strncmp(token, "_netdev", 7) == 0)
+		return OPT_IGNORE;
 
 	return OPT_ERROR;
 }
@@ -850,20 +859,30 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 	int word_len;
 	int rc = 0;
 	int got_uid = 0;
+	int got_cruid = 0;
 	int got_gid = 0;
-	char user[32];
-	char group[32];
+	uid_t uid, cruid = 0;
+	gid_t gid;
+	char *ep;
+	struct passwd *pw;
+	struct group *gr;
+	/*
+	 * max 32-bit uint in decimal is 4294967295 which is 10 chars wide
+	 * +1 for NULL, and +1 for good measure
+	 */
+	char txtbuf[12];
 
 	/* make sure we're starting from beginning */
 	out[0] = '\0';
 
 	/* BB fixme check for separator override BB */
-	if (getuid()) {
+	uid = getuid();
+	if (uid != 0)
 		got_uid = 1;
-		snprintf(user, sizeof(user), "%u", getuid());
+
+	gid = getgid();
+	if (gid != 0)
 		got_gid = 1;
-		snprintf(group, sizeof(group), "%u", getgid());
-	}
 
 	if (!data)
 		return EX_USAGE;
@@ -952,10 +971,12 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 					"target ip address argument missing\n");
 			} else if (strnlen(value, MAX_ADDRESS_LEN) <=
 				MAX_ADDRESS_LEN) {
+				strcpy(parsed_info->addrlist, value);
 				if (parsed_info->verboseflag)
 					fprintf(stderr,
 						"ip address %s override specified\n",
 						value);
+				goto nocopy;
 			} else {
 				fprintf(stderr, "ip address too long\n");
 				return EX_USAGE;
@@ -1006,43 +1027,59 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 			break;
 
 		case OPT_UID:
-			if (value && *value) {
-				got_uid = 1;
-				if (!isdigit(*value)) {
-					struct passwd *pw;
+			if (!value || !*value)
+				goto nocopy;
 
-					if (!(pw = getpwnam(value))) {
-						fprintf(stderr,
-							"bad user name \"%s\"\n",
-							value);
-						return EX_USAGE;
-					}
-					snprintf(user, sizeof(user), "%u",
-						 pw->pw_uid);
-				}
-				else
-					strlcpy(user, value, sizeof(user));
+			got_uid = 1;
+			errno = 0;
+			uid = strtoul(value, &ep, 10);
+			if (errno == 0)
+				goto nocopy;
+
+			pw = getpwnam(value);
+			if (pw == NULL) {
+				fprintf(stderr, "bad user name \"%s\"\n", value);
+				return EX_USAGE;
 			}
+
+			uid = pw->pw_uid;
+			goto nocopy;
+
+		case OPT_CRUID:
+			if (!value || !*value)
+				goto nocopy;
+
+			got_cruid = 1;
+			errno = 0;
+			cruid = strtoul(value, &ep, 10);
+			if (errno == 0)
+				goto nocopy;
+
+			pw = getpwnam(value);
+			if (pw == NULL) {
+				fprintf(stderr, "bad user name \"%s\"\n", value);
+				return EX_USAGE;
+			}
+			cruid = pw->pw_uid;
 			goto nocopy;
 
 		case OPT_GID:
-			if (value && *value) {
-				got_gid = 1;
-				if (!isdigit(*value)) {
-					struct group *gr;
+			if (!value || !*value)
+				goto nocopy;
 
-					if (!(gr = getgrnam(value))) {
-						fprintf(stderr,
-							"bad group name \"%s\"\n",
-							value);
-						return EX_USAGE;
-					}
-					snprintf(group, sizeof(group), "%u",
-						 gr->gr_gid);
-				}
-				else
-					strlcpy(group, value, sizeof(group));
+			got_gid = 1;
+			errno = 0;
+			gid = strtoul(value, &ep, 10);
+			if (errno == 0)
+				goto nocopy;
+
+			gr = getgrnam(value);
+			if (gr == NULL) {
+				fprintf(stderr, "bad group name \"%s\"\n", value);
+				return EX_USAGE;
 			}
+
+			gid = gr->gr_gid;
 			goto nocopy;
 
 		/* fmask fall through to file_mode */
@@ -1104,6 +1141,12 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 		case OPT_NO_LOCK:
 			*filesys_flags &= ~MS_MANDLOCK;
 			break;
+		case OPT_MAND:
+			*filesys_flags |= MS_MANDLOCK;
+			goto nocopy;
+		case OPT_NOMAND:
+			*filesys_flags &= ~MS_MANDLOCK;
+			goto nocopy;
 		case OPT_DEV:
 			*filesys_flags &= ~MS_NODEV;
 			break;
@@ -1126,6 +1169,8 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 		case OPT_REMOUNT:
 			*filesys_flags |= MS_REMOUNT;
 			break;
+		case OPT_IGNORE:
+			goto nocopy;
 		}
 
 		/* check size before copying option to buffer */
@@ -1155,123 +1200,55 @@ nocopy:
 
 	/* special-case the uid and gid */
 	if (got_uid) {
-		word_len = strlen(user);
+		word_len = snprintf(txtbuf, sizeof(txtbuf), "%u", uid);
 
+		/* comma + "uid=" + terminating NULL == 6 */
 		if (out_len + word_len + 6 > MAX_OPTIONS_LEN) {
 			fprintf(stderr, "Options string too long\n");
 			return EX_USAGE;
 		}
 
 		if (out_len) {
-			strlcat(out, ",", out_len + word_len + 6);
+			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 5, "uid=%s", user);
+		snprintf(out + out_len, word_len + 5, "uid=%s", txtbuf);
 		out_len = strlen(out);
 	}
-	if (got_gid) {
-		word_len = strlen(group);
+	if (got_cruid) {
+		word_len = snprintf(txtbuf, sizeof(txtbuf), "%u", cruid);
 
-		if (out_len + 1 + word_len + 6 > MAX_OPTIONS_LEN) {
+		/* comma + "cruid=" + terminating NULL == 6 */
+		if (out_len + word_len + 8 > MAX_OPTIONS_LEN) {
 			fprintf(stderr, "Options string too long\n");
 			return EX_USAGE;
 		}
 
 		if (out_len) {
-			strlcat(out, ",", out_len + word_len + 6);
+			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 5, "gid=%s", group);
+		snprintf(out + out_len, word_len + 7, "cruid=%s", txtbuf);
 		out_len = strlen(out);
 	}
+	if (got_gid) {
+		word_len = snprintf(txtbuf, sizeof(txtbuf), "%u", gid);
 
+		/* comma + "gid=" + terminating NULL == 6 */
+		if (out_len + word_len + 6 > MAX_OPTIONS_LEN) {
+			fprintf(stderr, "Options string too long\n");
+			return EX_USAGE;
+		}
+
+		if (out_len) {
+			strlcat(out, ",", MAX_OPTIONS_LEN);
+			out_len++;
+		}
+		snprintf(out + out_len, word_len + 5, "gid=%s", txtbuf);
+	}
 	return 0;
 }
 
-/*
- * resolve "host" portion of parsed info to comma-separated list of
- * address(es)
- */
-static int resolve_host(struct parsed_mount_info *parsed_info)
-{
-	int rc;
-	/* 10 for max width of decimal scopeid */
-	char tmpbuf[NI_MAXHOST + 1 + 10 + 1];
-	const char *ipaddr;
-	size_t len;
-	struct addrinfo *addrlist, *addr;
-	struct sockaddr_in *sin;
-	struct sockaddr_in6 *sin6;
-
-	rc = getaddrinfo(parsed_info->host, NULL, NULL, &addrlist);
-	if (rc != 0) {
-		fprintf(stderr, "mount error: could not resolve address for "
-			"%s: %s\n", parsed_info->host,
-			rc == EAI_SYSTEM ? strerror(errno) : gai_strerror(rc));
-		/* FIXME: return better error based on rc? */
-		return EX_USAGE;
-	}
-
-	addr = addrlist;
-	while (addr) {
-		/* skip non-TCP entries */
-		if (addr->ai_socktype != SOCK_STREAM ||
-		    addr->ai_protocol != IPPROTO_TCP) {
-			addr = addr->ai_next;
-			continue;
-		}
-
-		switch (addr->ai_addr->sa_family) {
-		case AF_INET6:
-			sin6 = (struct sockaddr_in6 *)addr->ai_addr;
-			ipaddr = inet_ntop(AF_INET6, &sin6->sin6_addr, tmpbuf,
-					   sizeof(tmpbuf));
-			if (!ipaddr) {
-				rc = EX_SYSERR;
-				fprintf(stderr,
-					"mount error: problem parsing address "
-					"list: %s\n", strerror(errno));
-				goto resolve_host_out;
-			}
-
-			if (sin6->sin6_scope_id) {
-				len = strnlen(tmpbuf, sizeof(tmpbuf));
-				ipaddr = tmpbuf + len;
-				snprintf(tmpbuf, sizeof(tmpbuf) - len, "%%%u",
-					 sin6->sin6_scope_id);
-			}
-			break;
-		case AF_INET:
-			sin = (struct sockaddr_in *)addr->ai_addr;
-			ipaddr = inet_ntop(AF_INET, &sin->sin_addr, tmpbuf,
-					   sizeof(tmpbuf));
-			if (!ipaddr) {
-				rc = EX_SYSERR;
-				fprintf(stderr,
-					"mount error: problem parsing address "
-					"list: %s\n", strerror(errno));
-				goto resolve_host_out;
-			}
-
-			break;
-		default:
-			addr = addr->ai_next;
-			continue;
-		}
-
-		if (parsed_info->addrlist[0] != '\0')
-			strlcat(parsed_info->addrlist, ",",
-				sizeof(parsed_info->addrlist));
-		strlcat(parsed_info->addrlist, tmpbuf,
-			sizeof(parsed_info->addrlist));
-		addr = addr->ai_next;
-	}
-
-resolve_host_out:
-	freeaddrinfo(addrlist);
-	return rc;
-}
-
 static int parse_unc(const char *unc_name, struct parsed_mount_info *parsed_info)
 {
 	int length = strnlen(unc_name, MAX_UNC_LEN);
@@ -1535,16 +1512,16 @@ add_mtab(char *devname, char *mountpoint, unsigned long flags, const char *fstyp
 	mountent.mnt_freq = 0;
 	mountent.mnt_passno = 0;
 	rc = addmntent(pmntfile, &mountent);
+	if (rc) {
+		fprintf(stderr, "unable to add mount entry to mtab\n");
+		rc = EX_FILEIO;
+	}
 	endmntent(pmntfile);
 	unlock_mtab();
 	SAFE_FREE(mountent.mnt_opts);
 add_mtab_exit:
 	toggle_dac_capability(1, 0);
 	sigprocmask(SIG_SETMASK, &oldmask, NULL);
-	if (rc) {
-		fprintf(stderr, "unable to add mount entry to mtab\n");
-		rc = EX_FILEIO;
-	}
 
 	return rc;
 }
@@ -1626,10 +1603,22 @@ assemble_mountinfo(struct parsed_mount_info *parsed_info,
 	if (rc)
 		goto assemble_exit;
 
-	rc = resolve_host(parsed_info);
-	if (rc)
+	if (parsed_info->addrlist[0] == '\0')
+		rc = resolve_host(parsed_info->host, parsed_info->addrlist);
+
+	switch (rc) {
+	case EX_USAGE:
+		fprintf(stderr, "mount error: could not resolve address for "
+			"%s: %s\n", parsed_info->host,
+			rc == EAI_SYSTEM ? strerror(errno) : gai_strerror(rc));
 		goto assemble_exit;
 
+	case EX_SYSERR:
+		fprintf(stderr, "mount error: problem parsing address "
+			"list: %s\n", strerror(errno));
+		goto assemble_exit;
+	}
+
 	if (!parsed_info->got_user) {
 		/*
 		 * Note that the password will not be retrieved from the
@@ -1687,12 +1676,11 @@ int main(int argc, char **argv)
 	char *orgoptions = NULL;
 	char *mountpoint = NULL;
 	char *options = NULL;
-	char *dev_name = NULL, *orig_dev = NULL;
+	char *orig_dev = NULL;
 	char *currentaddress, *nextaddress;
 	int rc = 0;
 	int already_uppercased = 0;
 	size_t options_size = MAX_OPTIONS_LEN;
-	size_t dev_len;
 	struct parsed_mount_info *parsed_info = NULL;
 	pid_t pid;
 	const char *fstype;
@@ -1728,8 +1716,6 @@ int main(int argc, char **argv)
 		return EX_SYSERR;
 	}
 
-	parsed_info->flags = MS_MANDLOCK;
-
 	/* add sharename in opts string as unc= parm */
 	while ((c = getopt_long(argc, argv, "?fhno:rvVw",
 				longopts, NULL)) != -1) {
@@ -1839,25 +1825,6 @@ int main(int argc, char **argv)
 		goto mount_exit;
 	}
 
-	/* lengths of different strings + slashes + trailing \0 */
-	dev_len = strnlen(parsed_info->host, sizeof(parsed_info->host)) +
-	    strnlen(parsed_info->share, sizeof(parsed_info->share)) +
-	    strnlen(parsed_info->prefix, sizeof(parsed_info->prefix)) +
-	    2 + 1 + 1 + 1;
-	dev_name = calloc(dev_len, 1);
-	if (!dev_name) {
-		rc = EX_SYSERR;
-		goto mount_exit;
-	}
-
-	/* rebuild device name with forward slashes */
-	strlcpy(dev_name, "//", dev_len);
-	strlcat(dev_name, parsed_info->host, dev_len);
-	strlcat(dev_name, "/", dev_len);
-	strlcat(dev_name, parsed_info->share, dev_len);
-	strlcat(dev_name, "/", dev_len);
-	strlcat(dev_name, parsed_info->prefix, dev_len);
-
 	currentaddress = parsed_info->addrlist;
 	nextaddress = strchr(currentaddress, ',');
 	if (nextaddress)
@@ -1905,7 +1872,7 @@ mount_retry:
 	if (parsed_info->verboseflag)
 		fprintf(stderr, "\n");
 
-	rc = check_mtab(thisprogram, dev_name, mountpoint);
+	rc = check_mtab(thisprogram, orig_dev, mountpoint);
 	if (rc)
 		goto mount_exit;
 
@@ -1914,8 +1881,13 @@ mount_retry:
 	else
 		fstype = cifs_fstype;
 
-	if (!parsed_info->fakemnt
-	    && mount(dev_name, ".", fstype, parsed_info->flags, options)) {
+	if (!parsed_info->fakemnt) {
+		toggle_dac_capability(0, 1);
+		rc = mount(orig_dev, ".", fstype, parsed_info->flags, options);
+		toggle_dac_capability(0, 0);
+		if (rc == 0)
+			goto do_mtab;
+
 		switch (errno) {
 		case ECONNREFUSED:
 		case EHOSTUNREACH:
@@ -1950,15 +1922,15 @@ mount_retry:
 		goto mount_exit;
 	}
 
-	if (!parsed_info->nomtab)
-		rc = add_mtab(dev_name, mountpoint, parsed_info->flags, fstype);
+do_mtab:
+	if (!parsed_info->nomtab && !mtab_unusable())
+		rc = add_mtab(orig_dev, mountpoint, parsed_info->flags, fstype);
 
 mount_exit:
 	if (parsed_info) {
 		memset(parsed_info->password, 0, sizeof(parsed_info->password));
 		munmap(parsed_info, sizeof(*parsed_info));
 	}
-	SAFE_FREE(dev_name);
 	SAFE_FREE(options);
 	SAFE_FREE(orgoptions);
 	return rc;
